unit simopt;

{$mode objfpc}{$H+}

{$R+} //für dynamische Arrays

interface

uses
  Classes, SysUtils, Math, Grids, Dialogs,
  StdCtrls;

type

  TRequirementsResult = (rrTrue,rrFalse,rrLess,rrMore);

  TSimplexInf = record
    pivotCol, pivotRow : integer;
    tableau : integer;
    isSimplexSolution : boolean;
    isSolution : boolean;
    isFinished : boolean;
    hasNoSolution : boolean;
  end;

  TToolData = record
    BladeWidth : double;
  end;

  TCutListResult = record
    CutLength : double;
    RawLength : double;
    WasteLength : double;
    BarsCount : integer;
    RemainingLength : double;
  end;


  TMaterialData = class
  private
    FRawLength : double;
    FBeginLength : double;
    FEndLength : double;
    FDiameter : double;
    FCostPerKg : double;
    FLeftoverLength : double;
    function GetLength : double;
  public
    //constructor Create;
    //destructor Destroy;
    function GetWeightFromLength(length : double) : double;
  published
    property RawLength : double read FRawLength write FRawLength;
    property BeginLength : double read FBeginLength write FBeginLength;
    property EndLength : double read FEndLength write FEndLength;
    property Length : double read GetLength;
    property Diameter : double read FDiameter write FDiameter;
    property CostPerKg : double read FCostPerKg write FCostPerKg;
    property LeftoverLength : double read FLeftoverLength write FLeftoverLength;
  end;


  { TRequirementsList }

  TRequirementsListItem = record
    Amount : integer;
    CutLength : double;
  end;

  TRequirementsList = class
  private
  public
    list : array of TRequirementsListItem;

    function Count : integer;

    constructor Create;
    procedure Clear;
    procedure Add(item : TRequirementsListItem);
    procedure Output(StringGrid: TStringGrid);
  end;


  { TWasteCombination }

  TWasteCombinationArray = array of array of integer;

  TWasteCombination = class
  private
    FRequirementsList : TRequirementsList;
    procedure SetRequirementsList(AValue: TRequirementsList);
    procedure Add;
  public
    Combination : TWasteCombinationArray; //array of array of integer;
    CutLength, WasteLength: array of double;

    function Count : integer;

    constructor Create;
    procedure Clear;
    procedure Calc;
    procedure Output(StringGrid : TStringGrid);
    procedure SaveToFile(filename : string);
  published
    property RequirementsList : TRequirementsList read FRequirementsList write SetRequirementsList;
  end;



  { TCutList }
  TCutListType = (cltCurrent,cltOptimized);

  TCutListItem = record
    Amount : integer;
    AmountDestination : integer;
    CutLength : double;
  end;

  TCutListBar = record
    CombinationId : integer;
    Amount : integer;
    CutLength : double;
    WasteLength : double;
    RawLength : double;
  end;


  TCutList = class
  private
    FRemainingBarLength : double;
    function SumRaw: double;
  public
    Bars : array of TCutListBar;
    BarsItemAmount : array of array of integer;
    Items : array of TCutListItem;
    SimplexInf : TSimplexInf;

    function BarsCount : integer;
    function SumWaste : double;
    function SumCut : double;
    function GetCutListResult : TCutListResult;
    function DeltaLength : double;
    function GetCuttingSequence(id: integer): string;

    constructor Create;
    procedure Clear;
    procedure AddItems(RequirementsList : TRequirementsList);
    procedure AddBar(item : TCutListBar; WasteCombinationArray : TWasteCombinationArray; id : integer);
    procedure OutputItems(StringGrid: TStringGrid);
    procedure OutputBars(StringGrid: TStringGrid);
    procedure Assign(cutlist : TCutList);
    procedure UpdateItemsAmount;

  published
    property RemainingBarLength : double read FRemainingBarLength write FRemainingBarLength;
  end;

  { TWasteOptimization }

  TWasteOptimization = class
  private
    FRequirementsList : TRequirementsList;
    FCurrCutList : TCutList; //aktuelles Ergebnis des Tableaus
    FOptCutList : TCutList; //Bestes Ergebnis
    FRemainingBarLength : double;
    FTableau : integer;
    simTab,simTabNew : array of array of double;
    simResX,simResY : array of integer;
    pivotCol,pivotRow : integer;

    function CalcCurrCutListLength(var arr: TWasteCombinationArray): double;
    function FindPivotCol: integer;
    function FindPivotRow(col: integer): integer;
    function GetRequirementsList: TRequirementsList;
    function IsSimplexSolution: boolean;
    function MyDiv(a, b: double): double;
    function OptimizeTableau : integer;

    procedure ClearCurrCutListLength(var arr: TWasteCombinationArray);
    procedure SetRequirementsList(AValue: TRequirementsList);
  public
    WasteCombination : TWasteCombination;
    function RequirementsFullfilled: TRequirementsResult;
    function GetPivotStr: string;

    constructor Create;
    destructor Destroy;
    procedure Optimize;
    procedure OptimizeInit;
    function OptimizeStep : boolean;
    procedure Output(StringGrid : TStringGrid; rounded : boolean = false);
    procedure CalcResult;
  published
    property RequirementsList : TRequirementsList read GetRequirementsList write SetRequirementsList;
    property OptCutList : TCutList read FOptCutList;
    property CurrCutList : TCutList read FCurrCutList;
    property RemainingBarLength : double read FRemainingBarLength write FRemainingBarLength;
    property Tableau : integer read FTableau write FTableau;
  end;


  TProductionProgramItem = record
    CutLength : double;
    Date : string;
    Amount : integer;
    CutListResult : TCutListResult;
  end;

  { TProductionProgram }

  TProductionProgram = class
  private
    positionId, lastPositionId : integer;
    positionDate : string;
    FEOL : boolean;
    FRequirementsList : TRequirementsList;
    FWasteOptimization : TWasteOptimization;

  public
    Items : array of TProductionProgramItem;
    CutListResult : TCutListResult;
    function Count : integer;
    function GetDate : string;

    constructor Create;
    destructor Destroy;
    procedure Clear;
    procedure Add(item : TProductionProgramItem);
    procedure LoadFromFile(filename: string);
    procedure Output(StringGrid: TStringGrid);
    procedure OutputDates(const Strings : TStrings);
    procedure SetRequirementsListFromDate(datestr: string; rl: TRequirementsList
      );
    procedure Calc;
    procedure CalcInit;
    procedure CalcNext;
  published
    property EOL : boolean read FEOL;
    property WasteOptimization : TWasteOptimization read FWasteOptimization write FWasteOptimization;
    property RequirementsList : TRequirementsList read FRequirementsList write FRequirementsList;
  end;



  TOptimizationControlItem = record
    RawLength : double;
    CutListResult : TCutListResult;
  end;

  { TOptimizationControl }

  TOptimizationControl = class
  private
    FProductionProgram : TProductionProgram;
  public
    Items : array of TOptimizationControlItem;

    function Count : integer;

    constructor Create;
    destructor Destroy;
    procedure Clear;
    procedure Add(item : TOptimizationControlItem);
    procedure Output(StringGrid: TStringGrid);
    procedure SaveToFile(filename : string);
  published
    property ProductionProgram : TProductionProgram read FProductionProgram write FProductionProgram;

  end;


var
  MaterialData : TMaterialData;
  ToolData : TToolData;


function StrInt(s : string) : longint;
function StrFloat(s : string) : double;
procedure DecodeCSV(count : integer; s : string; var arr : array of string);


implementation

//uses main;

function StrInt(s : string) : longint;
var c : integer;
begin
  val(s,result,c);
end;

function StrFloat(s : string) : double;
begin
  try
    result:=StrToFloat(s);
  except
    result:=0;
  end;
end;

procedure DecodeCSV(count : integer; s : string; var arr : array of string);
var
  ii,jj,kk : integer;
  check : boolean;
  h : string;
begin
  h:=s+';';
  check:=true; jj:=0; kk:=1;
  for ii:=1 to length(h) do
    if check and (h[ii]=';') then
      begin
        if jj=count then check:=false
        else
          begin
            inc(jj);
            arr[jj]:=copy(h,kk,ii-kk);
            kk:=ii+1;
          end;
      end;
end;

{ TProductionProgram }

constructor TProductionProgram.Create;
begin
  inherited Create;

  FRequirementsList:=TRequirementsList.Create;
  FWasteOptimization:=TWasteOptimization.Create;
  FWasteOptimization.RequirementsList:=FRequirementsList;

  Clear;
end;

destructor TProductionProgram.Destroy;
begin
  FWasteOptimization.Free;
  FRequirementsList.Free;

  inherited Destroy;
end;

procedure TProductionProgram.Clear;
begin
  SetLength(Items,0);
end;

procedure TProductionProgram.LoadFromFile(filename : string);
var
  sl : TStringList;
  i : integer;
  item : TProductionProgramItem;
  ppRow : array[0..3] of string;

begin
  try
    sl:=TStringList.Create;
    sl.LoadFromFile(filename);
    for i:=1 to sl.Count-1 do
      begin
        DecodeCSV(3,sl[i],ppRow);
        item.Date:=ppRow[1];
        item.CutLength:=StrFloat(ppRow[2]);
        item.Amount:=StrInt(ppRow[3]);
        item.CutListResult.RawLength:=0;
        item.CutListResult.CutLength:=0;
        item.CutListResult.WasteLength:=0;
        item.CutListResult.BarsCount:=0;
        item.CutListResult.RemainingLength:=0;
        Add(item);
      end;

  finally
    sl.Free;
  end;

end;

procedure TProductionProgram.Add(item: TProductionProgramItem);
begin
  SetLength(Items,Length(Items)+1);
  Items[high(Items)]:=item;
end;

function TProductionProgram.Count: integer;
begin
  result:=Length(Items);
end;

procedure TProductionProgram.Output(StringGrid: TStringGrid);
var i,j : integer;
begin
  with StringGrid do begin
    ColCount:=7;
    if Length(Items)>1000 then j:=1000
    else j:=Length(Items);
    RowCount:=j+1;
    Cells[0,0]:='Datum';
    Cells[1,0]:='Sägelänge [mm]';
    Cells[2,0]:='Menge [Stk]';
    Cells[3,0]:='Schnitt [mm]';
    Cells[4,0]:='Verschnitt [mm]';
    Cells[5,0]:='Stäbe [Stk]';
    Cells[6,0]:='Restlänge [mm]';
    for i:=0 to j-1 do //high(Items)
      begin
        Cells[0,i+1]:=Items[i].Date;
        Cells[1,i+1]:=FloatToStr(Items[i].CutLength);
        Cells[2,i+1]:=IntToStr(Items[i].Amount);
        Cells[3,i+1]:=FloatToStr(Items[i].CutListResult.CutLength);
        Cells[4,i+1]:=FloatToStr(Items[i].CutListResult.WasteLength);
        Cells[5,i+1]:=IntToStr(Items[i].CutListResult.BarsCount);
        Cells[6,i+1]:=FloatToStr(Items[i].CutListResult.RemainingLength);
      end;
    FixedRows:=1;
    AutoSizeColumns;
  end;
end;

procedure TProductionProgram.OutputDates(const Strings: TStrings);
var
  i : integer;
  d : string;
begin
  Strings.Clear;
  d:='';
  for i:=0 to high(Items) do
    if d<>Items[i].Date then
      begin
        d:=Items[i].Date;
        Strings.Add(d);
      end;
end;

procedure TProductionProgram.CalcInit;
begin
  positionId:=0;
  lastPositionId:=0;
  FEOL:=false;
  CutListResult.RawLength:=0;
  CutListResult.CutLength:=0;
  CutListResult.WasteLength:=0;
  CutListResult.BarsCount:=0;
  CutListResult.RemainingLength:=0;
  WasteOptimization.RemainingBarLength:=MaterialData.Length;
end;

procedure TProductionProgram.CalcNext;
var
  r : TCutListResult;
  ritem : TRequirementsListItem;
begin
  positionDate:='';

  FRequirementsList.Clear;
  if positionId<=high(Items) then
    begin
      positionDate:=Items[positionId].Date;
      lastPositionId:=positionId;
    end;

  if positionDate<>'' then
    begin
      while (positionId<=high(Items)) and (positionDate=Items[positionId].Date) do begin
        if positionDate=Items[positionId].Date then begin
          ritem.Amount:=Items[positionId].Amount;
          ritem.CutLength:=Items[positionId].CutLength;
          FRequirementsList.Add(ritem);
          positionId:=positionId+1;
        end;
      end;
      if positionId=high(Items) then FEOL:=true
      else FEOL:=false;
    end
  else
    begin
      FEOL:=true;
    end;


  if FRequirementsList.Count=0 then exit;

  FWasteOptimization.Optimize;
  r:=FWasteOptimization.OptCutList.GetCutListResult;

  CutListResult.RawLength:=CutListResult.RawLength+r.RawLength;
  CutListResult.CutLength:=CutListResult.CutLength+r.CutLength;
  CutListResult.WasteLength:=CutListResult.WasteLength+r.WasteLength;
  CutListResult.BarsCount:=CutListResult.BarsCount+r.BarsCount;
  CutListResult.RemainingLength:=CutListResult.RemainingLength+r.RemainingLength;

  Items[lastPositionId].CutListResult:=r;
end;

procedure TProductionProgram.Calc;
begin
  CalcInit;
  repeat
    CalcNext;
  until FEOL=true;
end;


procedure TProductionProgram.SetRequirementsListFromDate(datestr : string; rl : TRequirementsList);
var
  ritem : TRequirementsListItem;
  i,p : integer;
begin
  p:=-1;
  for i:=0 to high(Items) do
    if Items[i].Date=datestr then
      begin
        p:=i;
        break;
      end;

  if p=-1 then exit;

  rl.Clear;

  while (p<=high(Items)) and (datestr=Items[p].Date) do begin
    if datestr=Items[p].Date then begin
      ritem.Amount:=Items[p].Amount;
      ritem.CutLength:=Items[p].CutLength;
      rl.Add(ritem);
      p:=p+1;
    end;
  end;
end;


function TProductionProgram.GetDate: string;
begin
  result:=Items[lastPositionId].Date;
end;


constructor TOptimizationControl.Create;
begin
  inherited Create;

  FProductionProgram:=TProductionProgram.Create;
  Clear;
end;

destructor TOptimizationControl.Destroy;
begin
  FProductionProgram.Free;

  inherited Destroy;
end;

procedure TOptimizationControl.Clear;
begin
  SetLength(Items,0);
end;

procedure TOptimizationControl.Add(item: TOptimizationControlItem);
var
  i : integer;
begin
  //Prüfen, ob Länge schon vorhanden ist
  for i:=0 to high(Items) do
    if item.RawLength=Items[i].RawLength then exit;

  SetLength(Items,Length(Items)+1);
  Items[high(Items)]:=item;
end;

function TOptimizationControl.Count: integer;
begin
  result:=Length(Items);
end;

procedure TOptimizationControl.Output(StringGrid: TStringGrid);
var i : integer;
begin
  with StringGrid do begin
    ColCount:=5;
    RowCount:=Length(Items)+1;
    Cells[0,0]:='Länge Rohmaterial [mm]';
    Cells[1,0]:='Stabanzahl [Stk]';
    Cells[2,0]:='Verbrauch Rohlängen [mm]';
    Cells[3,0]:='Schnitt [mm]';
    Cells[4,0]:='Verschnitt [mm]';
    for i:=0 to high(Items) do
      begin
        Cells[0,i+1]:=FloatToStr(Items[i].RawLength);
        Cells[1,i+1]:=IntToStr(Items[i].CutListResult.BarsCount);
        Cells[2,i+1]:=FloatToStr(Items[i].CutListResult.RawLength);
        Cells[3,i+1]:=FloatToStr(Items[i].CutListResult.CutLength);
        Cells[4,i+1]:=FloatToStr(Items[i].CutListResult.WasteLength);
      end;
    FixedRows:=1;
    AutoSizeColumns;
  end;
end;

procedure TOptimizationControl.SaveToFile(filename: string);
var
  i : integer;
  sl : TStringList;
begin
  try
    sl:=TStringList.Create;
    sl.Add('Länge Rohmaterial [mm];Stabanzahl [Stk];Verbrauch Rohlängen [mm];Schnitt [m];Verschnitt [m]');
    for i:=0 to high(Items) do
      begin
        sl.Add(FloatToStr(Items[i].RawLength)+';'
          +IntToStr(Items[i].CutListResult.BarsCount)+';'
          +FloatToStr(Items[i].CutListResult.RawLength)+';'
          +FloatToStr(Items[i].CutListResult.CutLength)+';'
          +FloatToStr(Items[i].CutListResult.WasteLength));
      end;
    sl.SaveToFile(filename);
  finally
    sl.Free;
  end;
end;


{ TWasteCombination }

procedure TWasteCombination.SetRequirementsList(AValue: TRequirementsList);
begin
  if FRequirementsList=AValue then Exit;
  FRequirementsList:=AValue;
  Clear;
end;

constructor TWasteCombination.Create;
begin
  inherited Create;
  Clear;
end;

procedure TWasteCombination.Clear;
var
  i : integer;
begin
  SetLength(CutLength,1);
  CutLength[0]:=0;

  SetLength(WasteLength,1);
  WasteLength[0]:=0;

  SetLength(combination,1);

  if FRequirementsList<>nil then
    begin
      SetLength(combination[0],FRequirementsList.Count+1);
      for i:=0 to FRequirementsList.Count do
        combination[0,i]:=0;
    end;

end;

procedure TWasteCombination.Calc;
var
  i,ii,j,jj,r : integer;
  finished,isTrue : boolean;
  sum,restwert,verbrauch : double;
  rest: array of double;

begin
  if FRequirementsList=nil then exit;

  Clear;
  SetLength(rest,FRequirementsList.Count+1);

  Add;
  for i:=1 to FRequirementsList.Count do
    begin
      rest[i]:=0;
      Combination[1,i]:=0;
    end;

  rest[0]:=MaterialData.Length;
  i:=1;
  j:=1;
  finished:=false;
  Combination[i,j]:=Trunc(rest[j-1]/(FRequirementsList.list[j].CutLength+ToolData.BladeWidth));
  repeat
    sum:=0;
    for jj:=1 to FRequirementsList.Count do
      sum:=sum+Combination[i,jj]*(FRequirementsList.list[jj].CutLength+ToolData.BladeWidth);
    rest[j]:=MaterialData.Length-sum;

    if rest[j]<(FRequirementsList.list[FRequirementsList.Count].CutLength+ToolData.BladeWidth) then
      begin
        for jj:=j+1 to FRequirementsList.Count do
          Combination[i,jj]:=0;
        restwert:=rest[j];
        verbrauch:=MaterialData.Length-rest[j];
        isTrue:=true;
        for jj:=1 to FRequirementsList.Count-1 do
            if Combination[i,jj]<>0 then
              begin
                isTrue:=false;
                break;
              end;
        if Combination[i,FRequirementsList.Count]=0 then isTrue:=false;
        if isTrue=true then finished:=true
        else
          begin
            CutLength[i]:=verbrauch;
            WasteLength[i]:=restwert;

            i:=i+1; Add;
            for jj:=1 to FRequirementsList.Count do
              begin
                Combination[i,jj]:=0;
                rest[jj]:=0;
              end;
            rest[0]:=MaterialData.Length;
            r:=FRequirementsList.Count-1;
            repeat
              if Combination[i-1,r]=0 then r:=r-1;
            until (Combination[i-1,r]<>0) or (r<=1);
            for jj:=1 to r-1 do
              Combination[i,jj]:=Combination[i-1,jj];
            Combination[i,r]:=Combination[i-1,r]-1;
            j:=r;
          end;
        end
    else
      begin
        j:=j+1;
        Combination[i,j]:=Trunc(rest[j-1]/(FRequirementsList.list[j].CutLength+ToolData.BladeWidth));
      end;
  until finished=true;

  CutLength[i]:=verbrauch;
  WasteLength[i]:=restwert;

end;

procedure TWasteCombination.Add;
begin
  if FRequirementsList=nil then exit;

  SetLength(combination,Length(combination)+1);
  SetLength(Combination[high(Combination)],FRequirementsList.Count+1);

  SetLength(CutLength,Length(CutLength)+1);
  SetLength(WasteLength,Length(WasteLength)+1);
end;

function TWasteCombination.Count: integer;
begin
  result:=high(combination);
end;

procedure TWasteCombination.Output(StringGrid: TStringGrid);
var
  i,j : integer;
begin
  if FRequirementsList=nil then exit;

  with StringGrid do begin
    BeginUpdate;
    ColCount:=FRequirementsList.Count+3;
    RowCount:=high(combination)+1;
    Cells[0,0]:='Kombination';
    Cells[1,0]:='Schnitt';
    Cells[2,0]:='Verschnitt';
    for i:=1 to FRequirementsList.Count do
      begin
        Cells[2+i,0]:=FloatToStr(FRequirementsList.list[i].CutLength);
        for j:=1 to high(combination) do
          Cells[2+i,j]:=IntToStr(combination[j,i]);
      end;
    for j:=1 to high(combination) do
      begin
        Cells[0,j]:=IntToStr(j);
        Cells[1,j]:=FloatToStr(CutLength[j]);
        Cells[2,j]:=FloatToStr(WasteLength[j]);
      end;
    FixedCols:=3;
    FixedRows:=1;
    AutoSizeColumns;
    EndUpdate;
  end;
end;

procedure TWasteCombination.SaveToFile(filename: string);
var
  sl : TStringList;
  i,j : integer;
begin
  try
    sl:=TStringList.Create;
   { for i:=1 to FRequirementsList.Count do
      begin
        Cells[2+i,0]:=FloatToStr(FRequirementsList.list[i].CutLength);
        for j:=1 to high(combination) do
          Cells[2+i,j]:=IntToStr(combination[j,i]);
      end;
    for j:=1 to high(combination) do
      begin
        Cells[0,j]:=IntToStr(j);
        Cells[1,j]:=FloatToStr(CutLength[j]);
        Cells[2,j]:=FloatToStr(WasteLength[j]);
      end;}

  finally
    sl.Free;
  end;
end;


{ TWasteOptimization }


constructor TWasteOptimization.Create;
begin
  inherited Create;

  WasteCombination:=TWasteCombination.Create;
  FRequirementsList:=nil;
  FCurrCutList:=TCutList.Create;
  FOptCutList:=TCutList.Create;
end;

destructor TWasteOptimization.Destroy;
begin
  FRequirementsList:=nil;
  WasteCombination.Free;
  FCurrCutList.Free;
  FOptCutList.Free;

  inherited Destroy;
end;

procedure TWasteOptimization.OptimizeInit;
var
  x,y : integer;
begin
  if FRequirementsList=nil then exit;

  FCurrCutList.Clear;
  FOptCutList.Clear;

  //Schnittkombinationen berechnen
  WasteCombination.Calc;


  //Array vorbereiten
  //ToDo: Achtung Arraybeginn bei 0 und high()
  SetLength(simTab,Length(WasteCombination.Combination)+1);
  for x:=0 to high(WasteCombination.Combination) do
    SetLength(simTab[x],Length(FRequirementsList.list)+1);

  //SetLength(simResX,Length(FRequirementsList.list)+1);
  SetLength(simResX,WasteCombination.Count+1);
  SetLength(simResY,Length(FRequirementsList.list)+1);

  simTab[0,0]:=0;

  for y:=1 to FRequirementsList.Count do
    begin
      simResY[y]:=y+WasteCombination.Count;
      simTab[0,y]:=FRequirementsList.list[y].Amount;
    end;
  for x:=1 to WasteCombination.Count do
    begin
      simResX[x]:=x;
      simTab[x,0]:=WasteCombination.CutLength[x]*(-1);
    end;

  for x:=1 to WasteCombination.Count do
    for y:=1 to FRequirementsList.Count do
      simTab[x,y]:=WasteCombination.Combination[x,y];

  SetLength(simTabNew,Length(simTab));
  for x:=0 to high(simTab) do
    SetLength(simTabNew[x],Length(simTab[x]));

  FTableau:=0;
end;


function TWasteOptimization.FindPivotCol : integer;
var
  x : integer;
  pivotMin : double;
begin
  result:=0;
  pivotMin:=0;
  //for x:=1 to WasteCombination.Count do
  for x:=WasteCombination.Count downto 1 do
    if simTab[x,0]<=pivotMin then
      begin
        pivotMin:=simTab[x,0];
        result:=x;
      end;
end;

function TWasteOptimization.FindPivotRow(col : integer) : integer;
var
  y : integer;
  baseMin : double;
  d : double;
begin
  result:=0;
  baseMin:=Double.MaxValue;
  for y:=1 to FRequirementsList.Count do
    if simTab[col,y]<>0 then
      begin
        d:=simTab[0,y]/simTab[col,y];
        if (d>0) and (d<baseMin) then
          begin
            baseMin:=d;
            result:=y;
          end;
      end;
end;

function TWasteOptimization.GetRequirementsList: TRequirementsList;
begin
  result:=WasteCombination.RequirementsList;
end;

function TWasteOptimization.MyDiv(a,b : double) : double;
begin
  if b=0 then result:=0
  else result:=a/b;
end;

function TWasteOptimization.IsSimplexSolution : boolean;
var i : integer;
begin
  for i:=1 to high(simResY) do
    if simResY[i]>high(simResX) then
      begin
        result:=false;
        exit;
      end;
  result:=true;
end;


function TWasteOptimization.OptimizeTableau: integer;
var
  x,y : integer;
  res : integer;
begin
  res:=1;

  for x:=1 to WasteCombination.Count do
    if simTab[x,0]<0 then res:=0;

  result:=res;

  if res=1 then exit; //Optimum gefunden
  if FTableau=1 then exit; //beim ersten Tableau nichts machen

  //Pivotspalte suchen
  pivotCol:=FindPivotCol;

  //Pivotzeile suchen
  pivotRow:=FindPivotRow(pivotCol);
  if pivotRow=0 then
    begin
      //es exisiert keine Lösung
      result:=2;
      exit;
    end;

  //Umformen der Tabelle
  //a) Pivotelement
  simTabNew[pivotCol,pivotRow]:=1/simTab[pivotCol,pivotRow];

  //b) Pivotspalte
  for y:=0 to FRequirementsList.Count do
    if y<>pivotRow then
      simTabNew[pivotCol,y]:=-simTab[pivotCol,y]/simTab[pivotCol,pivotRow];

  //c) Pivotzeile
  for x:=0 to WasteCombination.Count do
    if x<>pivotCol then
      simTabNew[x,pivotRow]:=simTab[x,pivotRow]/simTab[pivotCol,pivotRow];

  //d) alle Elemente
  for x:=0 to WasteCombination.Count do
    for y:=0 to FRequirementsList.Count do
      if (x<>pivotCol) and (y<>pivotRow) then
        simTabNew[x,y]:=simTab[x,y]-simTab[x,pivotRow]*MyDiv(simTab[pivotCol,y],simTab[pivotCol,pivotRow]);

  for x:=0 to WasteCombination.Count do
    for y:=0 to FRequirementsList.Count do
      simTab[x,y]:=simTabNew[x,y];

  //Tauschen der Indizes
  x:=simResY[pivotRow];
  simResY[pivotRow]:=simResX[pivotCol];
  simResX[pivotCol]:=x;
end;

procedure TWasteOptimization.SetRequirementsList(AValue: TRequirementsList);
begin
  if AValue=FRequirementsList then exit;
  WasteCombination.RequirementsList:=AValue;
  FRequirementsList:=AValue;
  FCurrCutList.AddItems(AValue);
end;

function TWasteOptimization.GetPivotStr : string;
begin
  result:='Pivot '+IntToStr(pivotCol)+'/'+IntToStr(pivotRow);
end;

function TWasteOptimization.RequirementsFullfilled : TRequirementsResult;
var
  y : integer;
begin
  result:=rrFalse;
  //if FRequirementsList=nil then exit;

  result:=rrTrue;
  //for y:=1 to FRequirementsList.Count do
    //if FCurrCutList.items[y-1].Amount<FRequirementsList.List[y].Amount then result:=false;

  for y:=0 To high(FCurrCutlist.Items) do
    if FCurrCutList.Items[y].Amount<FCurrCutList.Items[y].AmountDestination then result:=rrLess;

  for y:=0 To high(FCurrCutlist.Items) do
    if FCurrCutList.Items[y].Amount>FCurrCutList.Items[y].AmountDestination then result:=rrMore;
end;

function TWasteOptimization.CalcCurrCutListLength(var arr : TWasteCombinationArray) : double;
var
  y : integer;
  d1,d2 : integer;
  d3,d4,cut : double;
begin
  SetLength(arr,1);
  SetLength(arr[0],Length(FCurrCutList.Items)+1);
  arr[0,0]:=0;
  cut:=0;
  for y:=0 to high(FCurrCutList.Items) do
    begin;
      d1:=FCurrCutList.Items[y].AmountDestination;
      d2:=FCurrCutList.Items[y].Amount;
      arr[0,y+1]:=d1-d2;

      //negative Werte nicht zulassen
      if arr[0,y+1]<0 then arr[0,y+1]:=0;

      d3:=FCurrCutList.Items[y].CutLength;
      d4:=arr[0,y+1]*(d3+ToolData.BladeWidth);
      cut:=cut+d4;
    end;
  result:=cut;
end;

procedure TWasteOptimization.ClearCurrCutListLength(var arr : TWasteCombinationArray);
var y : integer;
begin
  SetLength(arr,1);
  SetLength(arr[0],Length(FCurrCutList.Items)+1);
  for y:=0 to high(arr[0]) do
    arr[0,y]:=0;
end;

procedure TWasteOptimization.CalcResult;
var
  y,x : integer;
  b : boolean;
  cut,waste,rbl,cutBar : double;
  bar : TCutListBar;
  arr : TWasteCombinationArray;
  d1,d2 : integer;

begin
  if FRequirementsList=nil then exit;

  FCurrCutList.Clear;
  FCurrCutList.AddItems(FRequirementsList);

  for y:=1 to FRequirementsList.Count do
    if (simResY[y]<=WasteCombination.Count) and (simResY[y]>0) then
    begin
      //aufrunden: Ceil() kann interessant sein, wenn Mehrmengen produziert werden dürfen
      //abrunden:  Floor()
      bar.Amount:=Floor(simTab[0,y]);
      if bar.Amount>0 then
        begin
          bar.CombinationId:=simResY[y];
          bar.WasteLength:=WasteCombination.WasteLength[bar.CombinationId]
            +MaterialData.BeginLength+MaterialData.EndLength;
          bar.CutLength:=WasteCombination.CutLength[bar.CombinationId];
          bar.RawLength:=MaterialData.RawLength;
          FCurrCutList.AddBar(bar,WasteCombination.Combination,bar.CombinationId);
        end;
    end;


  if RequirementsFullfilled=rrTrue then
    begin
      FCurrCutList.SimplexInf.isSolution:=true;
      exit; //geforderte Mengen wurden erreicht
    end;


  //die exakten Mengen der geforderten Lösung wurden nicht erreicht
  //Reststab aufbrauchen
  FCurrCutList.UpdateItemsAmount;
  cut:=CalcCurrCutListLength(arr);

  if cut>FRemainingBarLength then
    begin
      //BEST-FIT:
      //Beste nächste Lösung aus den Kombinationen suchen

      d2:=0;
      repeat
        d2:=d2+1;
        waste:=MaterialData.RawLength;
        d1:=0; //beste Id

        for x:=1 to high(WasteCombination.Combination) do
          begin
            b:=true;
            for y:=0 to high(FCurrCutList.Items) do
              if WasteCombination.Combination[x,y+1]>FCurrCutList.Items[y].AmountDestination-FCurrCutList.Items[y].Amount then b:=false;
            if b=true then
              begin
                if WasteCombination.WasteLength[x]<waste then
                  begin
                    d1:=x; //beste Id
                    waste:=WasteCombination.WasteLength[x];
                  end;
              end;
          end;

        if d1>0 then
          begin
            bar.Amount:=1;
            bar.CombinationId:=d1;
            bar.WasteLength:=WasteCombination.WasteLength[bar.CombinationId]
              +MaterialData.BeginLength+MaterialData.EndLength;
            bar.CutLength:=WasteCombination.CutLength[bar.CombinationId];
            bar.RawLength:=MaterialData.RawLength;
            FCurrCutList.AddBar(bar,WasteCombination.Combination,bar.CombinationId);
          end;

        until (d1=0) or (d2>999);

        //reicht Restlänge jetzt aus?
        cut:=CalcCurrCutListLength(arr);
        if cut<=FRemainingBarLength then
          begin
            bar.Amount:=1;
            bar.CombinationId:=0;
            bar.CutLength:=cut;
            //bar.RawLength:=FRemainingBarLength;
            rbl:=FRemainingBarLength-MaterialData.BeginLength;

            //kann Reststück wiederverwendet werden?
            if rbl-cut<MaterialData.LeftoverLength then
              begin
                //alles Verschnitt, da Restlänge zu klein
                bar.RawLength:=FRemainingBarLength;
                bar.WasteLength:=FRemainingBarLength-cut+MaterialData.BeginLength;
                FCurrCutList.RemainingBarLength:=0;
              end
            else
              begin
                //kein Verschnitt, da Restlänge wiederverwendet wird
                bar.RawLength:=cut+MaterialData.BeginLength;
                bar.WasteLength:=MaterialData.BeginLength;
                FCurrCutList.RemainingBarLength:=rbl-cut;
              end;
            FCurrCutList.AddBar(bar,arr,0);
          end
        else
          begin
            //DIRECT-CUT:
            //Restlänge reicht nicht aus und sonst keine Lösung gefunden.
            //also so sägen wie es kommt

            //Delta-Mengen einzeln durchgehen und soviele Stäbe erzeugen wie benötigt
            //beginnen mit dem Reststab, am Ende bleibt ein neuer Reststab übrig

            ClearCurrCutListLength(arr);

            if FRemainingBarLength=0 then cutBar:=MaterialData.RawLength
            else cutBar:=FRemainingBarLength; //mit Reststab beginnen

            FCurrCutList.UpdateItemsAmount;
            cut:=0;
            for y:=0 to high(FCurrCutList.Items) do
              for x:=1 to FCurrCutList.Items[y].AmountDestination-FCurrCutList.Items[y].Amount do
                begin
                  if cut+FCurrCutList.Items[y].CutLength+ToolData.BladeWidth>cutBar-MaterialData.EndLength then
                    begin
                      //Länge übersteigt den aktuellen Stab
                      //Stab hinzufügen und neuen ganzen Stab beginnen
                      bar.Amount:=1;
                      bar.CombinationId:=-1;
                      bar.CutLength:=cut;
                      bar.RawLength:=cutBar;
                      bar.WasteLength:=cutBar-cut;
                      FCurrCutList.AddBar(bar,arr,0);
                      cutBar:=MaterialData.RawLength;
                      cut:=0;

                      ClearCurrCutListLength(arr);
                    end;
                  cut:=cut+FCurrCutList.Items[y].CutLength+ToolData.BladeWidth;
                  arr[0,y+1]:=arr[0,y+1]+1;
                end;

            bar.Amount:=1;
            bar.CombinationId:=-2;
            bar.CutLength:=cut;
            bar.RawLength:=cutBar;
            bar.WasteLength:=cutBar-cut+MaterialData.EndLength+MaterialData.BeginLength;

            //kann Reststück wiederverwendet werden?
            if cutBar-cut<MaterialData.LeftoverLength then
              begin
                //alles Verschnitt, da Restlänge zu klein
                bar.WasteLength:=FRemainingBarLength-cut;
                FCurrCutList.RemainingBarLength:=0;
              end
            else
              begin
                //kein Verschnitt, da Restlänge wiederverwendet wird
                bar.WasteLength:=MaterialData.BeginLength;
                FCurrCutList.RemainingBarLength:=cutBar-cut;
              end;
            FCurrCutList.AddBar(bar,arr,0);

          end;

    end
  else
    begin
      //Restlänge reicht aus
      bar.Amount:=1;
      bar.CombinationId:=0;
      bar.CutLength:=cut;
      rbl:=FRemainingBarLength-MaterialData.BeginLength;

      //kann Reststück wiederverwendet werden?
      if rbl-cut<MaterialData.LeftoverLength then
        begin
          //alles Verschnitt, da Restlänge zu klein
          bar.RawLength:=FRemainingBarLength;
          bar.WasteLength:=FRemainingBarLength-cut;
          FCurrCutList.RemainingBarLength:=0;
        end
      else
      begin
        //kein Verschnitt, da Restlänge wiederverwendet wird
        bar.RawLength:=cut+MaterialData.BeginLength;
        bar.WasteLength:=0;
        FCurrCutList.RemainingBarLength:=rbl-cut;
      end;
      FCurrCutList.AddBar(bar,arr,0);
    end;

  if RequirementsFullfilled=rrTrue then
    begin
      //geforderte Mengen wurden erreicht
      FCurrCutList.SimplexInf.isSolution:=true;
    end;
end;


procedure TWasteOptimization.Optimize;
var
  finished,cancel : boolean;
begin
  if FRequirementsList=nil then exit;
  if FRequirementsList.Count=0 then
    begin
      FCurrCutList.SimplexInf.hasNoSolution:=true;
      exit;
    end;

  cancel:=false;
  OptimizeInit;
  repeat
    finished:=OptimizeStep;
    if FTableau>100 then cancel:=true;
  until (finished=true) or (cancel=true);
  if cancel=true then FCurrCutList.SimplexInf.hasNoSolution:=true;
end;


function TWasteOptimization.OptimizeStep : boolean;
var
  res : integer;
  cut,rbl : double;
  bar : TCutListBar;
  arr : TWasteCombinationArray;

begin
  if FRequirementsList=nil then exit;

  //Wenn Requirementlist <= Stablänge, dann keine Optimierung
  FCurrCutList.Clear;
  FCurrCutList.AddItems(FRequirementsList);

  cut:=CalcCurrCutListLength(arr);

  if cut<=FRemainingBarLength then
    begin
      //Restlänge reicht aus
      bar.Amount:=1;
      bar.CombinationId:=0;
      bar.CutLength:=cut;
      bar.RawLength:=FRemainingBarLength;
      rbl:=FRemainingBarLength-MaterialData.BeginLength;

      //kann Reststück wiederverwendet werden?
      if rbl-cut<MaterialData.LeftoverLength then
        begin
          //alles Verschnitt, da Restlänge zu klein
          bar.WasteLength:=FRemainingBarLength-cut;
          FCurrCutList.RemainingBarLength:=0;
        end
      else
      begin
        //kein Verschnitt, da Restlänge wiederverwendet wird
        bar.WasteLength:=MaterialData.BeginLength;
        FCurrCutList.RemainingBarLength:=rbl-cut;
      end;
      FCurrCutList.AddBar(bar,arr,0);
      FCurrCutList.SimplexInf.isSolution:=true;
      FCurrCutList.SimplexInf.isFinished:=true;
      FOptCUtList.Assign(FCurrCUtList);
      exit;
    end;

  //Tableau's berechnen

  res:=OptimizeTableau;
  CalcResult;

  FCurrCutList.Simplexinf.isSimplexSolution:=IsSimplexSolution;

  if res>0 then FCurrCutList.SimplexInf.isFinished:=true;
  if res=2 then FCurrCutList.SimplexInf.hasNoSolution:=true;

  FCurrCutList.SimplexInf.pivotCol:=pivotCol;
  FCurrCutList.SimplexInf.pivotRow:=pivotRow;
  FCurrCutList.SimplexInf.tableau:=FTableau;

  if FCurrCutList.SimplexInf.isSolution=true then
    begin
      if FOptCutList.SimplexInf.isSolution=false then //bisher gab es noch keine Lösung
        FOptCutList.Assign(FCurrCutList)
      else
        begin
          //Prüfen, ob besser als aktuelle Lösung
          if FCurrCutList.SumWaste<FOptCutList.SumWaste then FOptCutList.Assign(FCurrCutList);
        end;
    end;

  if FCurrCutList.SimplexInf.isFinished=false then FTableau:=FTableau+1;

  result:=FCurrCutList.SimplexInf.isFinished;
end;

procedure TWasteOptimization.Output(StringGrid: TStringGrid; rounded: boolean = false);
var x,y : integer;
begin
  with StringGrid do begin
    RowCount:=FRequirementsList.Count+2;
    ColCount:=WasteCombination.Count+2;
    FixedCols:=2;
    FixedRows:=2;

    if rounded then Cells[0,0]:=FormatFloat('0.###',simTab[0,0])
    else Cells[0,0]:=FloatToStr(simTab[0,0]);

    for x:=1 to WasteCombination.Count do
      begin
        Cells[1+x,0]:='x'+IntToStr(simResX[x]);
        if rounded then Cells[1+x,1]:=FormatFloat('0.###',simTab[x,0])
        else Cells[1+x,1]:=FloatToStr(simTab[x,0]);
      end;
    for y:=1 to FRequirementsList.Count do
      begin
        Cells[0,1+y]:='x'+IntToStr(simResY[y]);
        if rounded then Cells[1,1+y]:=FormatFloat('0.###',simTab[0,y])
        else Cells[1,1+y]:=FloatToStr(simTab[0,y]);
      end;
    for x:=1 to WasteCombination.Count do
      for y:=1 to FRequirementsList.Count do
        if rounded then Cells[1+x,1+y]:=FormatFloat('0.###',simTab[x,y])
        else Cells[1+x,1+y]:=FloatToStr(simTab[x,y]);
    AutoSizeColumns;
  end;
end;

{ TRequirementsList }

constructor TRequirementsList.Create;
begin
  inherited Create;
  Clear;
end;

procedure TRequirementsList.Clear;
begin
  SetLength(list,1);
  list[0].CutLength:=0;
  list[0].Amount:=0;
end;

procedure TRequirementsList.Add(item: TRequirementsListItem);
var
  i : integer;
  found : boolean;
begin
  //Prüfen, ob Sägelänge schon vorhanden ist
  found:=false;
  for i:=0 to high(list) do
    if item.CutLength=list[i].CutLength then
      begin
        found:=true;
        list[i].Amount:=list[i].Amount+item.Amount;
        break;
      end;

  if found=false then
    begin
      //Todo: List aufsteigend sortieren nach CutLength
      SetLength(list,length(list)+1);
      list[high(list)]:=item;
    end;
end;

function TRequirementsList.Count: integer;
begin
  result:=high(list);       //oder +1???
end;

procedure TRequirementsList.Output(StringGrid: TStringGrid);
var i : integer;
begin
  with StringGrid do begin
    ColCount:=2;
    RowCount:=high(list)+1;
    Cells[0,0]:='Länge [mm]';
    Cells[1,0]:='Menge [Stk]';
    for i:=1 to high(list) do
      begin
        Cells[0,i]:=FloatToStr(list[i].CutLength);
        Cells[1,i]:=IntToStr(list[i].Amount);
      end;
    //FixedCols:=1;
    FixedRows:=1;
  end;
end;


{ TCutList }

constructor TCutList.Create;
begin
  inherited Create;
  Clear;
end;

procedure TCutList.Clear;
begin
  FRemainingBarLength:=0;

  SetLength(Items,0);
  SetLength(Bars,0);
  SetLength(BarsItemAmount,0);

  SimplexInf.isSolution:=false;
  SimplexInf.isFinished:=false;
  SimplexInf.hasNoSolution:=false;
end;

procedure TCutList.AddBar(item : TCutListBar; WasteCombinationArray : TWasteCombinationArray; id : integer);
var
  i,j,k : integer;
  found : boolean;
begin
  //Gibt es den Rohmaterial-Zuschnitt schon? Dann nur Menge erhöhen

  for i:=0 to high(Bars) do
    if (Bars[i].CombinationId=item.CombinationId) and (Bars[i].RawLength=item.RawLength) then
      begin
        found:=true;

        for j:=0 to high(BarsItemAmount[i]) do
          if WasteCombinationArray[id,j+1]<>BarsItemAmount[i,j] then
            begin
              found:=false;
              break;
            end;
        if found=true then
          begin
            //Menge erhöhen
            Bars[i].Amount:=Bars[i].Amount+1;
            UpdateItemsAmount;
            exit;
          end;
      end;


  SetLength(Bars,Length(Bars)+1);
  Bars[high(Bars)]:=item;

  SetLength(BarsItemAmount,Length(BarsItemAmount)+1);

  SetLength(BarsItemAmount[high(BarsItemAmount)],
    Length(WasteCombinationArray[id])-1); //Nummer startet bei 1
  j:=high(WasteCombinationArray[id]);
  for i:=1 to j do
    begin
      k:=WasteCombinationArray[id,i];
      BarsItemAmount[high(BarsItemAmount),i-1]:=k;
    end;
  UpdateItemsAmount;
end;

procedure TCutList.UpdateItemsAmount;
var i,j : integer;
begin
  for i:=0 to high(Items) do
    Items[i].Amount:=0;

  for i:=0 to high(Bars) do
    for j:=0 to high(BarsItemAmount[i]) do
      Items[j].Amount:=Items[j].Amount+BarsItemAmount[i,j]*Bars[i].Amount;
end;


procedure TCutList.AddItems(RequirementsList : TRequirementsList);
var i : integer;
begin
  if RequirementsList=nil then
    begin
      SetLength(Items,0);
    end
  else
    begin
      SetLength(Items,Length(RequirementsList.list)-1); //beginnt bei 1
      for i:=1 to high(RequirementsList.list) do
        begin
          Items[i-1].AmountDestination:=RequirementsList.list[i].Amount;
          Items[i-1].CutLength:=RequirementsList.list[i].CutLength;
          Items[i-1].Amount:=0;
        end;
    end;
end;

procedure TCutList.OutputItems(StringGrid: TStringGrid);
var i : integer;
begin
  UpdateItemsAmount;
  with StringGrid do begin
    ColCount:=4;
    RowCount:=high(Items)+2;
    Cells[0,0]:='Länge [mm]';
    Cells[1,0]:='Soll [Stk]';
    Cells[2,0]:='Erreicht [Stk]';
    Cells[3,0]:='Delta [Stk]';
    for i:=0 to high(Items) do
      begin
        Cells[0,i+1]:=FloatToStr(Items[i].CutLength);
        Cells[1,i+1]:=IntToStr(Items[i].AmountDestination);
        Cells[2,i+1]:=IntToStr(Items[i].Amount);
        Cells[3,i+1]:=IntToStr(Items[i].AmountDestination-Items[i].Amount);
      end;
    FixedCols:=1;
    FixedRows:=1;
    AutoSizeColumns;
  end;
end;

function TCutList.GetCuttingSequence(id: integer): string;
var
  i,j : integer;
  s : string;
begin
  //Schnittfolgezusammenstellen
  s:='';

  for i:=0 to high(BarsItemAmount[id]) do
    begin
      j:=BarsItemAmount[id,i];
      if j>0 then s:=s+IntToStr(j)+'•'+FloatToStr(Items[i].CutLength)+', ';
    end;

  if copy(s,length(s)-1,2)=', ' then s:=copy(s,1,length(s)-2);

  result:=s;
end;

procedure TCutList.OutputBars(StringGrid: TStringGrid);
var i : integer;
begin
  with StringGrid do begin
    ColCount:=5;
    RowCount:=high(Bars)+3;
    Cells[0,0]:='Rohlänge [mm]';
    Cells[1,0]:='Menge [Stk]';
    Cells[2,0]:='Verschnitt [mm]';
    Cells[3,0]:='Kombination';
    Cells[4,0]:='Schnittfolge';

    Cells[0,1]:='Summe';
    Cells[1,1]:=IntToStr(BarsCount);
    Cells[2,1]:=FloatToStr(SumWaste);
    Cells[3,1]:='';
    Cells[4,1]:='';

    for i:=0 to high(Bars) do
      begin
        Cells[0,i+2]:=FloatToStr(Bars[i].RawLength);
        Cells[1,i+2]:=IntToStr(Bars[i].Amount);
        Cells[2,i+2]:=FloatToStr(Bars[i].WasteLength);
        Cells[3,i+2]:=IntToStr(Bars[i].CombinationId);
        Cells[4,i+2]:=GetCuttingSequence(i);
      end;
    FixedCols:=1;
    FixedRows:=2;
    AutoSizeColumns;
  end;
end;

procedure TCutList.Assign(cutlist: TCutList);
var i,j : integer;
begin
  SetLength(Bars, Length(cutlist.Bars));
  for i:=0 to high(cutlist.Bars) do
    Bars[i]:=cutlist.Bars[i];

  SetLength(BarsItemAmount, Length(cutlist.BarsItemAmount));
  for i:=0 to high(cutlist.BarsItemAmount) do
    SetLength(BarsItemAmount[i],Length(cutlist.BarsItemAmount[i]));
  for i:=0 to high(BarsItemAmount) do
    for j:=0 to high(BarsItemAmount[0]) do
      BarsItemAmount[i,j]:=cutlist.BarsItemAmount[i,j];

  SetLength(Items, Length(cutlist.Items));
  for i:=0 to high(cutlist.Items) do
    Items[i]:=cutlist.Items[i];

  SimplexInf:=cutlist.SimplexInf;
  FRemainingBarLength:=cutlist.RemainingBarLength;
end;

function TCutList.BarsCount: integer;
var
  i,b : integer;
begin
  b:=0;
  for i:=0 to high(Bars) do
    b:=b+Bars[i].Amount;
  result:=b;
end;

function TCutList.SumWaste: double;
var
  i : integer;
  waste : double;
begin
  waste:=0;
  for i:=0 to high(Bars) do
    waste:=waste+Bars[i].WasteLength*Bars[i].Amount;
  result:=waste;
end;

function TCutList.SumCut: double;
var
  i : integer;
  cut : double;
begin
  cut:=0;
  for i:=0 to high(Bars) do
    cut:=cut+Bars[i].CutLength*Bars[i].Amount;
  result:=cut;
end;

function TCutList.SumRaw: double;
var
  i : integer;
  raw : double;
begin
  raw:=0;
  for i:=0 to high(Bars) do
    raw:=raw+Bars[i].RawLength*Bars[i].Amount;
  result:=raw;
end;

function TCutList.GetCutListResult: TCutListResult;
begin
  result.CutLength:=SumCut;
  result.WasteLength:=SumWaste;
  result.BarsCount:=BarsCount;
  result.RemainingLength:=FRemainingBarLength;
  result.RawLength:=SumRaw;
end;

function TCutList.DeltaLength: double;
var
  i : integer;
  cut : double;
begin
  UpdateItemsAmount;
  cut:=0;
  for i:=0 to high(Items) do
    cut:=cut+(Items[i].AmountDestination-Items[i].Amount)*Items[i].CutLength;
  result:=cut;
end;


function TMaterialData.GetLength: double;
begin
  result:=FRawLength-FBeginLength-FEndLength;
end;

function TMaterialData.GetWeightFromLength(length: double): double;
begin
  result:=pi*sqr((FDiameter/2))*7.85/1000*length;
end;




initialization
  MaterialData:=TMaterialData.Create;


finalization
  MaterialData.Free;


end.

